package fig.servlet;

import fig.basic.*;
import java.util.*;
import java.text.*;
import java.io.*;

/**
 * Substitutes the value of a field based on some replacements.
 */
public class ValueProcessor {
  static interface Processor {
    public String process(String s);
  }
  static class ReplaceProcessor implements Processor {
    private final String a, b;
    private final boolean global;
    public ReplaceProcessor(String a, String b, boolean global) {
      this.a = a; this.b = b;
      this.global = global;
    }
    public String process(String s) {
      if(s == null) return null;
      if(global) return s.replaceAll(a, b);
      else return s.replaceFirst(a, b);
    }
    public String toString() { return String.format("s/%s/%s/%s", a, b, global?"g":""); }
  }
  static class DoubleFormatProcessor implements Processor {
    public String process(String s) {
      double x = Utils.parseDoubleEasy(s);
      if(Double.isNaN(x)) return s;
      return Fmt.D(x);
    }
    public String toString() { return "DOUBLE"; }
  }
  static class DateFormatProcessor implements Processor {
    private static final DateFormat standardDateFormat =
          new SimpleDateFormat("EEE MMM dd kk:mm:ss zzz yyyy");

    public String process(String s) {
      try {
        Date d = standardDateFormat.parse(s);
        return Fmt.formatEasyDateTime(d.getTime());
      } catch(ParseException e) {
        return s;
      }
    }
    public String toString() { return "DATE"; }
  }
  static class BytesFormatProcessor implements Processor {
    public String process(String s) {
      try {
        return Fmt.bytesToString(Long.parseLong(s));
      } catch(NumberFormatException e) {
        return s;
      }
    }
  }


  private List<Processor> processors;

  // Tabs separate 
  // Format: s/true/EM\ts/false/var
  public ValueProcessor(String description) {
    this.processors = new ArrayList();
    if(description == null) return;
    for(String subStr : StrUtils.split(description, "\t")) {
      if(subStr.equals("DOUBLE"))
        processors.add(new DoubleFormatProcessor());
      else if(subStr.equals("DATE"))
        processors.add(new DateFormatProcessor());
      else if(subStr.equals("BYTES"))
        processors.add(new BytesFormatProcessor());
      else {
        // Replace: s/a/b/g
        subStr = subStr.replaceAll("\\\\/", "__SLASH__"); // Don't break on escaped slashes
        String[] tokens = StrUtils.split(subStr, "/");
        for (int i = 0; i < tokens.length; i++)
          tokens[i] = tokens[i].replaceAll("__SLASH__", "/");
        if(tokens.length == 0 || !tokens[0].equals("s")) continue;
        if(tokens.length == 2) tokens = new String[] {tokens[0], tokens[1], "", ""};
        if(tokens.length == 3) tokens = new String[] {tokens[0], tokens[1], tokens[2], ""};
        if(tokens.length != 4) continue;
        processors.add(new ReplaceProcessor(tokens[1], tokens[2], tokens[3].equals("g")));
      }
    }
  }

  public String process(String s) {
    if(s == null) return null;
    for(Processor processor : processors)
      s = processor.process(s);
    return s;
  }

  public boolean isIdentity() { return processors.size() == 0; }
  public String toString() { return StrUtils.join(processors, "\t"); }
}
